/* -----------------------------------------------------------------------------
 * d.cxx
 *
 *     D wrapper module.
 *
 * Author(s) : Andy Friesen
 *
 * Copyright (C) 2004
 * See the file LICENSE for information on usage and redistribution.
 * ----------------------------------------------------------------------------- */

char cvsroot_d_cxx[] = "$Header: /cvsroot/SWIG/Source/Modules/d.cxx,v 1.50 2003/12/23 21:45:38 cheetah Exp $";

#include <limits.h> // for INT_MAX
#include <ctype.h>
#include "swigmod.h"
#ifndef MACSWIG
#   include "swigconfig.h"
#endif

#include <cassert>
#include <string>
#include <vector>
#include <map>

namespace {

    // HACK: unified standin for operator delete:
    // Specialize for whatever types have to be destroyed differently.
    template <typename T>
    void destroy(T* t);
    
    template<>
    void destroy<DOH>(DOH* doh) { Delete(doh); }

    template<>
    void destroy<Wrapper>(Wrapper* w) { DelWrapper(w); }

    struct ScopedPtrBase { };

    template <class T>
    struct ScopedPtr : ScopedPtrBase {
        ScopedPtr(T* t)  : _data(t) { }
        ScopedPtr()      : _data(0) { }

        ScopedPtr(ScopedPtr& rhs) {
            (*this) = rhs;
        }

        virtual ~ScopedPtr() { 
            if (_data != 0) {
                destroy(_data);
            }
        }

        T* operator ->() const { return  _data; }
        T* get()         const { return  _data; }
     
        ScopedPtr& operator = (T* t) {
            destroy(_data);
            _data = t;
            return *this;
        }

        ScopedPtr& operator = (ScopedPtr& rhs) {
            (*this) = rhs;
            rhs = 0;
            return *this;
        }

    private:
        T* _data;
        //ScopedPtr(ScopedPtr&) { } // disallow copying
    };

    typedef ScopedPtr<String> ScopedString;

    template <typename T>
    ScopedPtr<T> makeAutoPtr(T* t) {
        return ScopedPtr<T>(t);
    }
}

// Evil little macro for automatically destroying a local when its scope dies.
// magic: Assigning a temporary to a const& extends the life of the temporary
// to that of the reference.
// NOTE: this is dumb and pointless.  Kill.
#define AUTO_HIDDEN(x, y) x ## y
#define AUTO_HIDE2(x, y) AUTO_HIDDEN(x, y)
#define AUTO(x) ScopedPtrBase const& AUTO_HIDE2(auto_secret, __LINE__) = makeAutoPtr(x);

/**
 * D language wrapper
 * TODO: deal with nested types.
 * FIXME: unnamed parameters in the source text cause incorrect output in
 * the output C++ code.
 */
class D : public Language {

    /// Context for functions: saves deduction effort.
    /// Stored in the node as its funcType property
    /// TODO: is there a better way?
    enum FuncType {
        ft_global       = (1 << 0), // is not a member of a class
        ft_static       = (1 << 1), // is a static member of a class
        ft_method       = (1 << 2), // is a member of a class
        ft_overload     = (1 << 3), // has overloads
        ft_ctor         = (1 << 4), // is a ctor
        ft_dtor         = (1 << 5),
        ft_attribute    = (1 << 6), // is an attribute
        ft_attribute_get= (1 << 7), // is an attribute get function
        ft_attribute_set= (1 << 8),

        // Composites for convenience
        ft_staticmethod = ft_static | ft_method,
    };

    virtual void main(int argc, char** argv) {
        SWIG_library_directory("d");
        Preprocessor_define("SWIGD 1", 0);
        SWIG_config_file("d.swg");
        SWIG_typemap_lang("d");

        allow_overloading();
    }

    virtual int top(Node* n) {
        String* moduleName  = Getattr(n, "name");         // module name: used for name decorations, effectively.

        if (moduleName == 0) {
            return fatalError("No module name specified.  Use the %module directive in your .i file.");
        }

        // Create output streams

        String* dshadowName = Getattr(n, "dshadow");        // Shadow classes that wrap the created C API.
        String* dmoduleName = Getattr(n, "dmoduleName");    // D module containing extern (C) functions.
        String* wrapperName = Getattr(n, "wrapper");        // extern "C" functions
        String* headerName = Getattr(n, "header");          // ... ?

        // fill in omitted defaults:
        if (dshadowName == 0) { dshadowName = NewStringf("%s.d", moduleName); }
        if (dmoduleName == 0) { dmoduleName = NewStringf("%s_extern.d", moduleName); }
        if (wrapperName == 0) { wrapperName = NewStringf("%s_wrap.cpp", moduleName); }

        if (headerName == 0) {
            // if no header, create a string buffer.
            // Merge into the wrapper file later.
            headerFile = NewString("");
        } else {
            // else create the file right here.
            headerFile = NewFile(headerName, "w");
        }

        wrapperFile = NewFile(wrapperName, "w");
        moduleFile = NewFile(dmoduleName, "w");
        shadowFile = NewFile(dshadowName, "w");

        // SWIG uses these names to insert code in various places.
        // Link them up with actual file handles.
        Swig_register_filebyname("wrapper", wrapperFile);
        Swig_register_filebyname("header", headerFile);
        Swig_register_filebyname("dextern", moduleFile);
        Swig_register_filebyname("dshadow", shadowFile);

        Printf(shadowFile, "private import %s_extern;\n\n", moduleName);

        Language::top(n);
        return SWIG_OK;
    }

    //
    virtual int enumDeclaration(Node* n) {
        Printf(stderr, "enum %s\n", Getattr(n, "name"));

        Printf(shadowFile, "enum %s {\n", Getattr(n, "name"));
        Language::enumDeclaration(n);
        Printf(shadowFile, "}\n");
        return SWIG_OK;
    }

    virtual int enumvalueDeclaration(Node* n) {
        Printf(shadowFile, "    %s = %s,\n", Getattr(n, "name"), Getattr(n, "enumvalue"));
        return SWIG_OK;
    }

    virtual int globalfunctionHandler(Node* n) {
        Printf(stderr, "\tglobal function: %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_global);
        int result = Language::globalfunctionHandler(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    virtual int memberfunctionHandler(Node* n) {
        Printf(stderr, "\tmethod: %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_method);
        int result = Language::memberfunctionHandler(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    virtual int staticmemberfunctionHandler(Node* n) {
        Printf(stderr, "\tstatic method %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_method | ft_static);
        
        // skip the default static method handler: it mangles names in a bad way
        int result = Language::staticmemberfunctionHandler(n);
        SetInt(n, "funcType", flags);

        return result;
    }

    virtual int callbackfunctionHandler(Node* n) {
        return NYI(n);
    }

    //
    virtual int globalvariableHandler(Node *n) {
        Printf(stderr, "\tglobal attribute %s\n", Getattr(n, "name"));
        return Language::globalvariableHandler(n);
    }

    virtual int membervariableHandler(Node *n) {
        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_attribute);
        int result = Language::membervariableHandler(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    virtual int staticmembervariableHandler(Node *n) {
        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_static | ft_attribute);
        int result = Language::staticmembervariableHandler(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    //
    virtual int memberconstantHandler(Node* n) {
        Printf(stderr, "\tconstant attribute %s\n", Getattr(n, "name"));
        return Language::memberconstantHandler(n);
    }

    virtual int constructorHandler(Node* n) {
        Printf(stderr, "\tctor %s\n", Getattr(n, "name"));

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_ctor);
        int result = Language::constructorHandler(n);
        SetInt(n, "funcType", flags);
        return result;
    }

    virtual int copyconstructorHandler(Node* n) {
        Printf(stderr, "\tcopy ctor %s\n", Getattr(n, "name"));                

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_ctor);
        int result = Language::copyconstructorHandler(n);
        SetInt(n, "funcType", flags | ft_ctor);
        return result;
    }

    virtual int destructorHandler(Node* n) {
        Printf(stderr, "\tdtor\n");

        int flags = GetInt(n, "funcType");
        SetInt(n, "funcType", flags | ft_dtor);
        int result = Language::destructorHandler(n);
        SetInt(n, "funcType", flags | ft_dtor);
        return result;
    }

    // Maybe still too complex.  Simplify?
    virtual int classHandler(Node* n) {
        Printf(stderr, "\nclass %s:\n", Getattr(n, "name"));
        
        String* className = Getattr(n, "name");
        String* baseName = NewString("");
        bool hasBase = hasattr(n, "bases");

        if (hasBase) {
            List* baseList = Getattr(n, "bases");
            Iterator iter = First(baseList);
            if (iter.item) {
                Printv(baseName, Getattr(iter.item, "name"), NIL);
            }
            if (Next(iter).item != 0) {
                // As good a warning as any.
                Swig_warning(WARN_JAVA_MULTIPLE_INHERITANCE, input_file, line_number,
                    "Warning for %s proxy: Base %s ignored. Multiple inheritance "
                    "is not supported in D.\n", className, Getattr(iter.item,"name")
                );
            }
        }

        // impl pointer, protected constructor (used by all public constructors and subclasses)
        if (!hasBase) {
            Printv(shadowFile,
                "class ", className, " {\n",
                "\n",
                "protected void* swig_impl;\n",
                "protected bool  swig_own_ptr;\n",
                "protected this(void* ptr, bool own) {\n",
                "    swig_impl = ptr;\n",
                "    swig_own_ptr = own;\n",
                "}\n",
                "\n",
                NIL
            );
        } else {
            Printv(shadowFile,
                "class ", className, " : ", baseName, " {\n",
                "\n",
                "protected this(void* ptr, bool own) {\n",
                "    super(ptr, own);\n",
                "}\n",
                "\n",
                NIL
            );
        }

        Language::classHandler(n);

        Printf(shadowFile, "} // end class %s\n\n", className);

        Delete(baseName); // RIAA?!  C is dumb.

        return SWIG_OK;
    }

    //
    virtual int constantWrapper(Node* n) {
        String* varType = Getattr(n, "type");
        String* varName = Getattr(n, "name");
        String* varValue = quoteConstant(n);

        Printf(shadowFile, "const %s %s = %s;\n", varType, varName, varValue);

        Delete(varValue);

        return SWIG_OK;
    }

    virtual int functionWrapper(Node* node) {
        Setattr(node, "sym:name", mangleName(node));
        //Language::functionWrapper(node);
        int result;

        // C glue
        result = writeExternCFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // D glue
        result = writeExternDFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // wrapper interface
        result = writeWrapper(node);
        return result;
    }


    int writeExternCFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String* returnType = SwigType_str(type, 0);

        Wrapper* wrapper = NewWrapper();

        emit_attach_parmmaps(parms, wrapper);
        Swig_typemap_attach_parms("CtoExtern", parms, wrapper);

        Printf(wrapper->def, "DllExport %s CDecl %s(%s) {", returnType, name, ParmList_str(parms));

        emit_args(type, parms, wrapper);

        emit_action(node, wrapper);

        if (SwigType_type(type) != T_VOID) {
            Printf(wrapper->code, "return result;");
        }

        Printf(wrapper->code, "}");
        Wrapper_print(wrapper, wrapperFile);

        destroy(wrapper);

        // Is this correct? SWIG's refcounting stuff is unclear to me.
        destroy(returnType); 

        return SWIG_OK;
    }

    int writeExternDFunc(Node* node) {
        String*    name  = Getattr(node,"sym:name");
        SwigType*  type  = Getattr(node,"type");
        ParmList*  parms = Getattr(node,"parms");

        String*    argList = NewString("");
        
        String* returnType = SwigType_str(type, 0);

        Swig_typemap_attach_parms("CtoD", parms, 0);

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node, "", 0);

        int parmCount = emit_num_arguments(parms);

        // Construct argument list: transform each argument type
        // and make up a name.
        Parm* parm = parms;
        for (int i = 0; i < parmCount; i++) {
            String* mappedType = Getattr(parm, "tmap:CtoD");

            String* oldType = SwigType_str(Getattr(parm, "type"), 0);

            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                Printf(stderr, "wrapExternDFunc: Can't convert argument type %s in function %s\n", oldType, name);
                Printv(argList, oldType, NIL);
            }

            Printf(argList, " arg%i", i + 1);

            if (i < parmCount - 1) {
                Printf(argList, ", ");
            }

            parm = nextSibling(parm);
        }

        if (mappedReturnType == 0) {
            mappedReturnType = returnType;
        }

        Printf(moduleFile, "extern (C) %s %s(%s);\n", mappedReturnType, name, argList);

        destroy(argList);

        return SWIG_OK;
    }

    int writeWrapper(Node* node) {
        // Now the D wrapper

        if (!hasattr(node, "funcType")) {
            Printf(stderr, "Internal error on function %s in D::writeWrapper.  Maybe?\n", Getattr(node, "name"));
            return SWIG_ERROR;
        }

        int flags = GetInt(node, "funcType");

        bool isAttribute = bool(flags & ft_attribute);
        bool isCtor = bool(flags & ft_ctor);
        bool isDtor = bool(flags & ft_dtor);
        bool isMethod = bool(flags & ft_method);
        bool isStatic = bool(flags & ft_static);
        bool isGlobal = bool(flags & ft_global);

        int result = SWIG_ERROR;
        if (isCtor) {
            result = writeWrapperCtor(node);

        } else if (isDtor) {
            result = writeWrapperDtor(node);

        } else if (isMethod || isStatic || isGlobal || isAttribute) {
            result = writeWrapperFunc(node, isGlobal, isStatic);

        } else {
            Printf(stderr, "Skipping %s %s: D needs no wrapper for pure C things.\n", 
                Getattr(node, "nodeType"), Getattr(node, "name")
            );

            result = SWIG_OK;
        }

        return result;
    }

    int writeWrapperFunc(Node* node, bool isGlobal, bool isStatic) {
        String*   name       = Getattr(node, "name");       // happy pretty name
        String*   wrapName   = Getattr(node, "sym:name");   // mangled name
        ParmList* parms      = Getattr(node, "parms");
        Node*     returnType = Getattr(node, "type");
        bool      hasThis    = !isGlobal && !isStatic;

        // chop everything before ::
        if (char* pos = Strstr(name, "::")) {
            char* lastPos;
            while (pos) {
                lastPos = pos;
                pos = Strstr(pos + 1, "::");
            }
            name = NewString(lastPos + 2);
        }

        Wrapper* wrapper = NewWrapper();

        Swig_typemap_attach_parms("CtoD", parms, wrapper);

        // string-ized argument list for the C function
        String* argumentList = NewString("");
        // argument signature for the shadow method
        String* signature = NewString("");

        // Generate argument conversion and signature
        Node* parm = parms;

        // 'self' argument gets special treatment
        if (hasThis) {
            assert(parm != 0);

            String* selfName = Getattr(parms, "name");

            Wrapper_new_local(wrapper, selfName, "void* self");
            Printf(wrapper->code, "%s = swig_impl;\n", selfName);

            Printf(argumentList, "%s", selfName);

            // move over, print comma in argumentList if there is a non-self argument
            parm = nextSibling(parm);
            if (parm != 0) { 
                Printf(argumentList, ", ");
            }
        }

        while (parm != 0) {
            SwigType* argType = Getattr(parm, "type");
            String* name = Getattr(parm, "name");
            String* lname = Getattr(parm, "lname");
            SwigType* mappedType = Getattr(parm, "tmap:CtoD");
            Setattr(parm, "ltype", mappedType);

            argType = SwigType_str(argType, 0);

            String* argConvert = Swig_typemap_lookup("Dout", mappedType, name, lname, "$source", "$target", wrapper);
            Printf(stderr, "\t%s\n\t%s\n\t%s\n\n", argType, mappedType, argConvert);

            String* decl = NewStringf("%s %s", mappedType, lname);
            
            Wrapper_new_local(wrapper, lname, decl);
            destroy(decl);

            Printf(wrapper->code, "%s\n", argConvert);
            Printf(argumentList, "%s", lname);

            Printf(signature, "%s %s", mappedType, name);

            parm = nextSibling(parm);
            if (parm != 0) {
                Printv(argumentList, ", ", NIL);
                Printv(signature, ", ", NIL);
            }
        }

        if (isStatic) {
            Printf(wrapper->def, "static ");
        }

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node, "", wrapper);
        Printf(wrapper->def, "%s %s(%s) {\n", mappedReturnType, name, signature);

        String* action = NewStringf("%s(%s)", wrapName, argumentList);
        String* a1 = Swig_typemap_lookup("Dargout", returnType, name, action, "$result", "$target", 0);
        Replaceall(a1, "$result", "result");
        Printv(wrapper->code, a1, "\n", NIL);

        if (SwigType_type(returnType) != T_VOID) {
            String* returnTypeStr = SwigType_str(mappedReturnType, 0);
            String* decl = NewStringf("%s result", returnTypeStr);
            
            Wrapper_new_local(wrapper, returnTypeStr, decl);

            destroy(returnTypeStr);

            Printf(wrapper->code, "return result;");
        }

        Printf(wrapper->code, "}");

        Wrapper_print(wrapper, shadowFile);

        destroy(argumentList);
        destroy(signature);
        destroy(wrapper);
        return 0;
    }

    /// Returns a mangled name for the function
    String* mangleName(Node* node) {
        int flags = GetInt(node, "funcType");

        String* name = Getattr(node, "sym:name");
        String* result = NewStringf("SwigWrap_%s", name);

        if (hasattr(node, "sym:overloaded")) {
            // Append wart to disambiguate overloads
            Printv(result, Getattr(node, "sym:overname"), NIL);
        }

        return result;
    }

    int writeWrapperCtor(Node* node) {
        ParmList* parms = Getattr(node, "parms");
        String* name = Getattr(node, "name");
        String* wrapName = Getattr(node, "sym:name");

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        // argument names and types
        // ie double d1, double d2, ...
        String* argList = NewString("");
        AUTO(argList);

        // argument names, no types
        // ie d1, d2, ...
        String* callList = NewString("");

        int parmCount = emit_num_arguments(parms);
        Parm* iter = parms;
        for (int i = 0; i < parmCount; i++) {
            SwigType* parmType = Getattr(iter, "type");
            String* parmName = Getattr(iter, "name");

            String* mappedType = Getattr(iter, "tmap:DtoC");
            if (mappedType == 0) {
                String* oldType = SwigType_str(Getattr(iter, "type"), 0);
                Printf(stderr, "wrapCtor: Can't convert argument type %s in function %s\n", oldType, parmName);
                mappedType = oldType;
            }

            Printf(argList, "%s arg%i", mappedType, i + 1);
            Printf(callList, "arg%i", i + 1);

            if (i < parmCount - 1) {
                Printv(argList, ", ", NIL);
                Printv(callList, ", ", NIL);
            }

            iter = nextSibling(iter);
        }

        Printf(wrapper->def, "this(%s) {", argList);
        
        Printf(wrapper->code, "this(%s(%s), true);\n", wrapName, callList);
        Printf(wrapper->code, "}");

        Wrapper_print(wrapper, shadowFile);

        return SWIG_OK;
    }

    int writeWrapperDtor(Node* node) {
        // Base class defines protected swig_destroy() and ~this()
        // Subclasses just override swig_destroy()
        bool hasBase = hasattr(node, "bases");
        String* dtorName = Getattr(node, "sym:name");
        if (!hasBase) {
            // Root class: define swig_destroy method and actual destructor.
            Printf(shadowFile,
                "protected void swig_destroy()\n"
                "out {\n"
                "    assert(swig_impl === null);\n"
                "}\n"
                "body {\n"
                "    if (swig_own_ptr && swig_impl !== null) {\n"
                "        %s(swig_impl);\n"
                "    }\n"
                "    swig_impl = null;\n"
                "}\n"
                "\n"
                "~this() {\n"
                "    swig_destroy();\n"
                "    // HACK: The current DMD implementation will not enforce the out contract on\n"
                "    // swig_destroy if it is overridden\n"
                "    assert(swig_impl === null);\n"
                "}\n"
                "\n",
                dtorName
            );
        } else {
            // No actual destructor on subclasses:
            // Just override swig_destroy() method
            Printf(shadowFile, 
                "    protected override void swig_destroy() {\n"
                "        if (swig_own_ptr && swig_impl !== null) {\n"
                "            %s(swig_impl);\n"
                "        }\n"
                "        swig_impl = null;\n"
                "    }\n"
                "\n",
                dtorName
            );
        }
        return SWIG_OK;
    }

    /// Generates a D function call to an extern(C) proxy function
    /// FIXME?
    /// we forgot what this method was originally going to do.
    /// This function creates the whole shadow method now.
    String* dfunctionCall(Node* node) {
    }

    void dumpNode(Node* node) {
            for (Iterator iter = First(node); iter.item; iter = Next(iter)) {
            if (DohIsString(iter.item)) {
                Printf(stderr, "'%s'\t'%s'\n", iter.key, iter.item);
            } else {
                Printf(stderr, "'%s'\t<%s>\n", iter.key, Getattr(iter.item, "nodeType"));
            }
        }
        Printf(stderr, "\n\n\n");
    }

    int fatalError(String* s) {
        Printf(stderr, "Error: %s\n");
        SWIG_exit(EXIT_FAILURE);
        return 0;
    }

    int NYI(Node* node) {
        Printf(stderr, "Not yet implemented: %s %s\n", Getattr(node, "nodeType"), Getattr(node, "name"));
        dumpNode(node);
        return SWIG_ERROR;
    }

    bool hasThis(Node* node) {
        if (!Getattr(node, "funcType")) {
            return false;
        }

        int i = GetInt(node, "funcType");
        if (i & (ft_static | ft_global)) {
            return false;
        } else {
            return true;
        }
    }

    bool hasattr(Node* node, String* key) {
        return Getattr(node, key) != 0;
    }

    void delattr(Node* node, String* key) {
        Setattr(node, key, 0);
    }

    // Puts the quotes on string/char literals
    // (SWIG strips them for some reason)
    String* quoteConstant(Node* value) {
        int valueType = SwigType_type(Getattr(value, "type"));
        if (valueType == T_STRING || valueType == T_CHAR) {
            char quote = valueType == T_STRING ? '\"' : '\'';
            String* oldValue = Getattr(value, "value");

            return NewStringf("%c%s%c", quote, oldValue, quote);
        } else {
            return Copy(Getattr(value, "value"));
        }
    }

    // attributes:

    /// Contains the implementation of the wrapped C API
    File* wrapperFile;

    /// TODO: compile flag to enable this, as opposed to merging it with wrapperFile
    File* headerFile;

    /// D source file that declares extern(C) functions for the contents of wrapperFile (above)
    File* moduleFile;

    /// Wraps the C API in a delicious, candy coated interface.
    File* shadowFile;

    /// Not used. (not useful?  Provide anyway for the sake of compatibility with other backends?)
#if 0
    String* c_init;   // init code. (also goes into wrapperFile, I think)
#endif

    static const char* usage;
};   /* class D */

extern "C" Language *
swig_d(void) {
  return new D();
}

// Why does this have to go out here?  C++ is dumb.
const char *D::usage = "\
D Options (available with -d)\n\
     -package <name> - set name of the package to <name>\n\
\n";
